#pragma once

#include <Base_Provider.h>
#include <unordered_map>
#include <librdkafka/rdkafkacpp.h>
#include <atomic>
#include <thread>
#include <Base_Configurator.h>

/**
 * @brief The namespace inaf::oasbo::Providers contains classes related to providers.
 */
namespace inaf::oasbo::Providers {

/**
 * @brief The KafkaProvider class is a derived class of BaseProvider and provides functionality for sendig packets to Kafka.
 * BaseProvider.h for more information.
 */
class KafkaProvider: public BaseProvider {
protected:

	RdKafka::Producer *producer = nullptr; /**< Pointer to the Kafka producer object. */
	std::atomic<bool> stopFlag = false; /**< Atomic flag to control the polling thread. */
	std::thread pollThread; /**< Thread for polling Kafka events. */
	class DeliveryReportCb;
	DeliveryReportCb *dr_cb = nullptr; /**< Pointer to the delivery report callback object. */

	void pollingThreadFunction(); /**< Function executed by the polling thread. */

	KafkaProvider(); /**< Default constructor. */
	KafkaProvider(std::string ip, int port, std::string topic); /**< Constructor with parameters. */

public:

	std::string brokerIp; /**< IP address of the Kafka broker. */
	int brokerPort; /**< Port number of the Kafka broker. */

	void setDest(std::string dest) override;

	std::string getDest() override;

	int write(Packets::BasePacket&) override;

	int write(Packets::BasePacket&, std::string dest) override;

	int close() override;

	int open() override;

	bool isOpen() override;

	/**
	 * @brief Destructor.
	 */
	~KafkaProvider();

	friend class KafkaProviderBuilder;

};

/**
 * @brief The KafkaProviderBuilder class is responsible for building KafkaProvider objects.
 */
class KafkaProviderBuilder {
protected:
	KafkaProvider *provider; /**< Pointer to the KafkaProvider object being built. */

public:
	std::string config_target { "kafkaprovider" }; /**< Configuration target for the Kafka provider. */
	std::string ip_key { "ip" }; /**< Configuration key for the IP address. */
	std::string port_key { "port" }; /**< Configuration key for the port number. */
	std::string topic_key { "topic" }; /**< Configuration key for the topic. */

	/**
	 * @brief Default constructor.
	 */
	KafkaProviderBuilder();

	/**
	 * @brief Constructor with parameters.
	 * @param ip The IP address of the Kafka broker.
	 * @param port The port number of the Kafka broker.
	 * @param topic The topic to subscribe to.
	 */
	KafkaProviderBuilder(std::string ip, int port, std::string topic);

	/**
	 * @brief Destructor.
	 */
	~KafkaProviderBuilder();

	/**
	 * @brief Resets the builder to its initial state.
	 */
	void reset();

	/**
	 * @brief Configures the builder from a BaseConfigurator object.
	 * @param conf The BaseConfigurator object to configure from.
	 * @return A pointer to the KafkaProviderBuilder object.
	 */
	KafkaProviderBuilder* configFrom(Configurators::BaseConfigurator &conf);

	/**
	 * @brief Sets the IP address for the Kafka provider.
	 * @param ip The IP address to set.
	 * @return A pointer to the KafkaProviderBuilder object.
	 */
	KafkaProviderBuilder* setIp(std::string ip);

	/**
	 * @brief Sets the port number for the Kafka provider.
	 * @param port The port number to set.
	 * @return A pointer to the KafkaProviderBuilder object.
	 */
	KafkaProviderBuilder* setPort(int port);

	/**
	 * @brief Sets the topic for the Kafka provider.
	 * @param topic The topic to set.
	 * @return A pointer to the KafkaProviderBuilder object.
	 */
	KafkaProviderBuilder* setTopic(std::string topic);

	/**
	 * @brief Gets the built KafkaProvider object.
	 * @return A pointer to the KafkaProvider object.
	 */
	KafkaProvider* getProvider();
};
}
