#include <Kafka_Provider.h>
#include <memory>
#include <chrono>
#include <unistd.h>
#include <ctime>
#include <iomanip>

using namespace inaf::oasbo::Providers;

class KafkaProvider::DeliveryReportCb: public RdKafka::DeliveryReportCb {
public:
	void dr_cb(RdKafka::Message &message) {
		/* If message.err() is non-zero the message delivery failed permanently
		 * for the message. */
		if (message.err()) {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Kafka Provider]\t" << "Message delivery failed"
					<< std::endl;
		}
	}
};

KafkaProvider::KafkaProvider() :
		KafkaProvider("127.0.0.1", 9092, "Astri_ADAS_topic") {
}

KafkaProvider::KafkaProvider(std::string ip, int port, std::string topic_name) :
		brokerIp(ip), brokerPort(port) {
	setDest(topic_name);
}

void KafkaProvider::pollingThreadFunction() {
	time_t now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[Kafka Provider]\t" << "Polling thread start" << std::endl;
	while (!this->stopFlag) {
		if (isOpen())
			this->producer->poll(0);
		std::this_thread::sleep_for(std::chrono::seconds(1)); // pool every second
	}
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[Kafka Provider]\t" << "Polling thread stop" << std::endl;
}

int KafkaProvider::write(Packets::BasePacket &packet) {
	return KafkaProvider::write(packet, this->dest);
}

int KafkaProvider::write(Packets::BasePacket &packet, std::string dest) {
	if (!isOpen()) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[Kafka Provider]\t" << "Provider not open" << std::endl;
		return -1;
	}
	int attempts = 0;
	RdKafka::ErrorCode err;
	while (attempts < 5) { //try to deliver messagr 5 times, then report error.
		err = this->producer->produce(
		/* Topic name */
		dest,
		/* Partition:*/
		0,
		/* Make a copy of the value */
		RdKafka::Producer::RK_MSG_COPY /* Copy payload */,
		/* Value */
		(char*) packet.getPointerToMemory(),
				packet.getHeaderSize() + packet.getPayloadSize()
						+ packet.getTailSize(),
				NULL, 0, 0,
				NULL);
		if (err == RdKafka::ERR_NO_ERROR)
			return 1;

		attempts += 1;

	}
	time_t now = time(nullptr);
	std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[Kafka Provider]\t" << " Failed to produce to topic " << dest
			<< ": " << err << std::endl;
	return -1;
}

int KafkaProvider::open() {
	if (isOpen())
		close();
	RdKafka::Conf *conf = RdKafka::Conf::create(RdKafka::Conf::CONF_GLOBAL);
	std::string errstr;
	if (conf->set("bootstrap.servers",
			std::string(this->brokerIp).append(":").append(
					std::to_string(this->brokerPort)), errstr)
			!= RdKafka::Conf::CONF_OK) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[Kafka Provider]\t" << errstr << std::endl;
		delete conf;
		return -1;
	}
	this->dr_cb = new DeliveryReportCb();
	if (conf->set("dr_cb", dr_cb, errstr) != RdKafka::Conf::CONF_OK) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[Kafka Provider]\t" << errstr << std::endl;
		delete conf;
		return -1;
	}
	this->producer = RdKafka::Producer::create(conf, errstr);
	if (!this->producer) {
		this->producer = nullptr;
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[Kafka Provider]\t" << "Failed to create producer: "
				<< errstr << std::endl;
		delete conf;
		return -1;
	}

	delete conf;
	// Check if at least one broker is available
	RdKafka::Metadata *metadata;
	RdKafka::ErrorCode err = this->producer->metadata(true, NULL, &metadata,
			5000);
	if (err != RdKafka::ERR_NO_ERROR || metadata->brokers()->size() == 0) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[Kafka Provider]\t" << "No brokers available"
				<< std::endl;
		delete metadata;
		this->producer = nullptr;
		return -1;
	}
	delete metadata;
	this->stopFlag = false;
	this->pollThread = std::thread(&KafkaProvider::pollingThreadFunction, this); // start polling
	return 1;
}

int KafkaProvider::close() {
	if (!isOpen())
		return 1;
	time_t now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[Kafka Provider]\t" << "Flushing kafka queue in 10 seconds.."
			<< std::endl;
	producer->flush(10 * 1000);

	if (producer->outq_len() > 0) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[Kafka Provider]\t" << producer->outq_len()
				<< " message(s) were not delivered" << std::endl;
	}
	this->stopFlag = true; // Stopping the polling thread

	this->pollThread.join();

	delete producer;
	producer = nullptr;
	delete dr_cb;

	return 1;
}

bool KafkaProvider::isOpen() {
	return this->producer != nullptr;
}

void KafkaProvider::setDest(std::string dest){
	this->dest = dest;
}

std::string KafkaProvider::getDest(){
	return this->dest;
}


KafkaProvider::~KafkaProvider() {
	close();
}

